;-------------------------------------------------------------------------------------------------------------------------
; author		: Beno Gorancic
; last modified	: 05/06/2006
;
; most of the code used in this program has been adopted from code made available
; to students on UTS Online under the subject 48441 - Introductory Digital Systems
;-------------------------------------------------------------------------------------------------------------------------


		LIST      P=16F877, F=INHX8M
		#include <P16F877.inc>


;-------------------------------------------------------------------------------------------------------------------------
; CONFIGURATION WORD
;-------------------------------------------------------------------------------------------------------------------------
			; code protect off, debug off, program memory write protection off,
			; data EE memory code protection off, low voltage programming off,
			; brown out detection on, watch dog timer off,
			; high speed(>4MHz) xtal program write on
	    	__config _CP_OFF & _DEBUG_OFF & _WRT_ENABLE_OFF & _CPD_OFF & _LVP_OFF & _BODEN_ON & _WDT_OFF & _HS_OSC & _PWRTE_ON


;-------------------------------------------------------------------------------------------------------------------------
; DEFINITIONS
;-------------------------------------------------------------------------------------------------------------------------
#define		VALUE_CHANGED			0	; used to check if frequency/amplitude has changed
#define		SCALED_VALUE_CHANGED	1	; used to check if scaled frequency/amplitude has changed


;-------------------------------------------------------------------------------------------------------------------------
; MEMORY EQUATES
;-------------------------------------------------------------------------------------------------------------------------
Bank0Ram	equ	H'20'


;-------------------------------------------------------------------------------------------------------------------------
; MEMORY ALLOCATIONS
;-------------------------------------------------------------------------------------------------------------------------
	cblock	Bank0Ram					; beginning of access RAM
			Flags
			Temp			
			LCD_Temp_Buffer
			Axis_Temp_Value
			DelayA
			DelayB
			DelayC
			huns
			tens
			units
			Frequency
			Amplitude
			Old_Frequency
			Old_Amplitude
			Scaled_Frequency
			Scaled_Amplitude
			Old_Scaled_Frequency
			Old_Scaled_Amplitude
	endc


;-------------------------------------------------------------------------------------------------------------------------
; VECTORS
;-------------------------------------------------------------------------------------------------------------------------
	org		0x0000		; reset vector
	goto	MainLine

	org		0x0004		; interrupt vector
	goto	IntService


;-------------------------------------------------------------------------------------------------------------------------
; IntService - INTERRUPT SERVICE ROUTINE
;-------------------------------------------------------------------------------------------------------------------------
IntService		; empty routine, just in case an interrupt occurs
	retfie		; return from interrupt


;-------------------------------------------------------------------------------------------------------------------------
; Initialize - INITIALIZATION ROUTINE
;-------------------------------------------------------------------------------------------------------------------------
Initialize
	bcf		STATUS,RP0
	bcf		STATUS,RP1		; bank 0 selected
	clrf	PIR1
	clrf 	PIR2			; clear interrupt flags registers
	clrf 	PORTA
	clrf 	PORTB
	clrf 	PORTC
	clrf 	PORTD
	clrf 	PORTE			; clear ports to output all 0's
	clrf	INTCON			; ensure all interrupts are turned off
	clrf	Flags			; clear Flags
	bsf		RCSTA,SPEN		; serial port enabled
	bsf		STATUS,RP0		; set ram bank 1
	clrf	TRISA
	clrf	TRISB
	clrf	TRISC
	clrf	TRISD
	clrf	TRISE			; pins on ports A to E are outputs
	clrf	PIE1			; dissable interrupts
	movlw 	H'3F'
	movwf	TRISA			; port A pins all inputs
	bsf		TRISC,6			; enable USART asynchronous transmit	
    movlw 	H'02'
	movwf 	ADCON1			; left justified, port E digital, port A analogue, Vref+ = Vdd, Vref- = Vss
	movlw	H'14'
	movwf	SPBRG			; Fosc = 20Mhz, BRGH = 1, baud rate = 57.6K,  we get: SPRGH value of D'20'
	bsf		TXSTA,TXEN		; transmission enabled
	bcf		TXSTA,SYNC		; asynchronous mode
	bsf		TXSTA,BRGH		; high speed baud rate
	bcf		STATUS,RP0		; set ram bank 0

	return


;-------------------------------------------------------------------------------------------------------------------------
; Startup_LCD - LCD STARTUP ROUTINE
;-------------------------------------------------------------------------------------------------------------------------
Startup_LCD
	bcf		STATUS,RP0
	bcf		STATUS,RP1		; set ram bank 0

	; the follwoing sequence is the recomemded power up sequence for an LCD as stated by the chipset manufacturer
	bcf		PORTE,0			; instruction register select
	movlw	.25
	call	Delay			; delay for LCD PowerUp
	movlw	.30
	call	LCDWrite8		; set LCD to 8 bits
	movlw	.5
	call	Delay			; delay for LCD
	movlw	.30
	call	LCDWrite8		; set LCD to 8 bits
	movlw	.5
	call	Delay			; delay for LCD
	movlw	.30
	call	LCDWrite8		; set LCD to 8 bits
	movlw	.5
	call	Delay			; delay for LCD
	movlw	.56
	call	LCDWrite8		; 8 bit interface 2 line, 5x7 character format
	movlw	.5
	call	Delay			; delay for LCD
	movlw	.56
	call	LCDWrite8		; 8 bit interface 2 line, 5x7 character format
	movlw	.5
	call	Delay			; delay for LCD

	; write user specific configuration data
	movlw 	H'6'
	call	LCDWrite8		; display shifting disabled, use cursor shifting instead
	movlw	.5
	call	Delay			; delay for LCD
	movlw	H'1'
	call	LCDWrite8		; clear all DDRAM and set cursor to pos 1
	movlw	.5
	call	Delay			; delay for LCD
	movlw	H'C'
	call	LCDWrite8		; display ON
	bsf		PORTE,0			; set to data register

	; write data to the LCD "10321747"
	movlw	H'31'
	call	LCDWrite8
	movlw	H'30'
	call	LCDWrite8
	movlw	H'33'
	call	LCDWrite8
	movlw	H'32'
	call	LCDWrite8
	movlw	H'31'
	call	LCDWrite8
	movlw	H'37'
	call	LCDWrite8
	movlw	H'34'
	call	LCDWrite8
	movlw	H'37'
	call	LCDWrite8

	; move cursor to 1st position 2nd line of the LCD
	bcf		PORTE,0			; set instruction register
	movlw	H'C0'
	call	LCDWrite8
	bsf		PORTE,0			; set data register

	; write data to the LCD "F="
	movlw	'F'
	call	LCDWrite8
	movlw	'='
	call	LCDWrite8

	; move cursor to 7th position 2nd line of the LCD
	bcf		PORTE,0			; set instruction register
	movlw	H'C6'
	call	LCDWrite8
	bsf		PORTE,0			; set data register

	; write data to the LCD "A="
	movlw	'A'
	call	LCDWrite8
	movlw	'='
	call	LCDWrite8

	return


;-------------------------------------------------------------------------------------------------------------------------
; MainLine - MAIN PROGRAM
;-------------------------------------------------------------------------------------------------------------------------
MainLine
	call    Initialize
	call	Startup_LCD
	call	Startup_HyperTerminal


;-------------------------------------------------------------------------------------------------------------------------
; MainLoop - MAIN PROGRAM LOOP
;-------------------------------------------------------------------------------------------------------------------------
MainLoop
	call	Analog_to_Digital
	call	Scale_Data
	call	Has_Data_Changed
	call	Has_Scaled_Data_Changed

	btfsc	Flags,VALUE_CHANGED
	call	Send_to_LCD
	bcf		Flags,VALUE_CHANGED

	btfsc	Flags,SCALED_VALUE_CHANGED
	call	Send_to_HyperTerminal
	bcf		Flags,SCALED_VALUE_CHANGED
	
	movfw	Scaled_Frequency
	movwf	Old_Scaled_Frequency
	movfw	Scaled_Amplitude
	movwf	Old_Scaled_Amplitude

	goto	MainLoop


;-------------------------------------------------------------------------------------------------------------------------
; Send_to_LCD - SENDS DATA TO THE LCD
;-------------------------------------------------------------------------------------------------------------------------
Send_to_LCD
	bcf		PORTE,0				; set instruction register
	movlw	H'C2'				; move cursor to 3rd position 2nd line of the LCD
	call	LCDWrite8

	movfw	Frequency
	call	Binary_to_BCD

	bsf		PORTE,0				; set data register
	movfw	huns
	call	LCDWrite8			; send hundreds
	movfw	tens
	call	LCDWrite8			; send tens
	movfw	units
	call	LCDWrite8			; send units

	bcf		PORTE,0				; set instruction register
	movlw	H'C8'				; move cursor to 9th position 2nd line of the LCD
	call	LCDWrite8

	movfw	Amplitude
	call	Binary_to_BCD

	bsf		PORTE,0				; set data register
	movfw	huns
	call	LCDWrite8			; send hundreds
	movfw	tens
	call	LCDWrite8			; send tens
	movfw	units
	call	LCDWrite8			; send units

	return


;-------------------------------------------------------------------------------------------------------------------------
; Has_Data_Changed - CHECKS IF FREQUENCY/AMPLITUDE CHANGED
;-------------------------------------------------------------------------------------------------------------------------
Has_Data_Changed
	movfw	Frequency				; move Frequency to W register
	subwf	Old_Frequency,W			; subtract Frequency from Old_Frequency and store in W
	btfss	STATUS,Z				; if W=0, STATUS=1, skip next operation
	bsf		Flags,VALUE_CHANGED

	movfw	Amplitude				; move Amplitude to W register
	subwf	Old_Amplitude,W			; subtract Amplitude from Old_Amplitude and store in W
	btfss	STATUS,Z				; if W=0, STATUS=1, skip next operation
	bsf		Flags,VALUE_CHANGED

	return


;-------------------------------------------------------------------------------------------------------------------------
; Binary_to_BCD - CONVERTS A BINARY NUMBER TO A BINARY-CODED-DECIMAL
;-------------------------------------------------------------------------------------------------------------------------
Binary_to_BCD
	clrf	huns		; clear hundreds
	clrf	tens		; clear tens
	clrf	units		; clear units

; do subtract hundred and counting while no borrow is generated
Loop_h
	incf	huns,F		; record one hun subtracted
	addlw	-D'100'		; subtract decimal hundred
	btfsc	STATUS,C	; if a borrow (C==0) then exit Loop_h
	goto	Loop_h
	decf	huns,F		; correct for one subtract
	addlw	D'100'		; add 100 to residue

; do subtact ten and counting while no borrow is generated
Loop_t
	incf	tens, F		; record one ten subtracted
	addlw	-D'10'		; subtract decimal ten
	btfsc	STATUS, C	; if a borrow (C==0) then exit Loop_t
	goto	Loop_t
	decf	tens, F		; correct for one subtract
	addlw	D'10'		; add 10 to residue
	movwf	units		; save the remainder in units

	movlw	H'30'		; hex ascii code for number 0
	addwf	huns,F		; add H'30' to get correct ascii value
	addwf	tens,F		; add H'30' to get correct ascii value
	addwf	units,F		; add H'30' to get correct ascii value

	return


;-------------------------------------------------------------------------------------------------------------------------
; Analog_to_Digital - ANALOG TO DIGITAL CONVERTER
;-------------------------------------------------------------------------------------------------------------------------
Analog_to_Digital
	movlw 	H'91'
	movwf	ADCON0			; Fosc/32, channel 2, GO/DONE 0, A/D converter on
	movlw	.1
	call	Delay			; delays 1ms before conversion
	bsf		ADCON0,GO_DONE	; ADC routine

Wait_Frequency
	btfsc	ADCON0,GO_DONE	; check if conversion is over
	goto	Wait_Frequency	; if not goto Wait_Frequency
	movfw	ADRESH			; move high byte to W
	movwf	Frequency		; move high byte to Frequency
	movlw	H'99'
	movwf	ADCON0			; Fosc/32, channel 3, GO/DONE 0, A/D converter on
	movlw	.1
	call	Delay			; delays 1ms before conversion
	bsf		ADCON0,GO_DONE	; ADC routine

Wait_Amplitude
	btfsc	ADCON0,GO_DONE	; check if conversion is over
	goto	Wait_Amplitude	; if not goto Wait_Amplitude
	movfw	ADRESH			; move high byte to W
	movwf	Amplitude		; move high byte to Amplitude

	return


;-------------------------------------------------------------------------------------------------------------------------
; LCDWrite8 - WRITE TO THE LCD IN 8 BIT MODE
;-------------------------------------------------------------------------------------------------------------------------
LCDWrite8
	movwf	LCD_Temp_Buffer		; temporally store data to be written to the LCD
	bcf		PORTE,2				; clear the enable pin
	movlw	.1
	call	Delay				; 1ms delay for LCD timing
	movf	LCD_Temp_Buffer,W
	movwf	PORTD				; write stored data to PORTD ie LCD data pins
	bsf		PORTE,2				; pulse enable pin
	nop
	bcf		PORTE,2
	return


;-------------------------------------------------------------------------------------------------------------------------
; Delay - DELAYS A SET AMOUNT OF ms; COUNTS CYCLES BASED ON 20MHZ CLOCK
;-------------------------------------------------------------------------------------------------------------------------
Delay
	movwf	DelayA
OuterLoop
	movlw	.29
	movwf	DelayB
MiddleLoop
	movlw	.42
	movwf	DelayC
InnerLoop
	nop
	decfsz	DelayC,F
	goto	InnerLoop
	decfsz	DelayB,F
	goto	MiddleLoop
	decfsz	DelayA,F
	goto	OuterLoop
	return


;-------------------------------------------------------------------------------------------------------------------------
; Startup_HyperTerminal - HYPER TERMINAL STARTUP ROUTINE
;-------------------------------------------------------------------------------------------------------------------------
Startup_HyperTerminal	
	bcf STATUS,RP0
	bcf STATUS,RP1				; select bank 0
	
	movlw	H'41'				; moves ascii char to W reg :	A
	call	Send_Char
	movlw	H'33'				;								3
	call	Send_Char
	movlw	H'20'				;								space
	call	Send_Char
	movlw	H'32'				;								2
	call	Send_Char
	movlw	H'30'				;								0
	call	Send_Char
	movlw	H'30'				;								0
	call	Send_Char
	movlw	H'36'				;								6
	call	Send_Char
	movlw	H'0D'				;								carriage return
	call	Send_Char
	movlw	H'0A'				;								new line
	call	Send_Char
	movlw	H'48'				;								H
	call	Send_Char
	movlw	H'28'				;								(
	call	Send_Char
	movlw	H'73'				;								s
	call	Send_Char
	movlw	H'29'				;								)
	call	Send_Char
	movlw	H'0D'				;								carriage return
	call	Send_Char
	movlw	H'0A'				;								new line
	call	Send_Char
	movlw	H'5E'				;								^
	call	Send_Char
	movlw	H'0D'				;								carriage return
	call	Send_Char
	movlw	H'0A'				;								new line
	call	Send_Char

	movlw	D'16'
	movwf	Axis_Temp_Value

; insert '|' 16 times
Vertical_Axis
	movlw	H'7C'				;								|
	call	Send_Char
	movlw	H'0D'				;								carriage return
	call	Send_Char
	movlw	H'0A'				;								new line
	call	Send_Char
	decfsz	Axis_Temp_Value		; decrement Axis_Temp_Value...
	goto	Vertical_Axis		; skip goto if 0

	movlw	D'65'
	movwf	Axis_Temp_Value

; insert '-' 65 times
Horizontal_Axis
	movlw	H'2D'				;								-
	call	Send_Char
	decfsz	Axis_Temp_Value		; decrement Axis_Temp_Value...
	goto	Horizontal_Axis		; skip goto if 0

	movlw	H'3E'				;								>
	call	Send_Char
	movlw	H'20'				;								space
	call	Send_Char
	movlw	H'66'				;								f
	call	Send_Char

	return


;-------------------------------------------------------------------------------------------------------------------------
; Send_Char - SENDS CHARACTER TO HYPER TERMINAL
;-------------------------------------------------------------------------------------------------------------------------
Send_Char
	movwf	TXREG			; move value from W register to transmit register
	nop
Check
	btfss	PIR1,TXIF		; checks if GO/DONE is set
	goto	Check			; if not goto Check

	return


;-------------------------------------------------------------------------------------------------------------------------
; Scale_Data - SCALES DOWN THE FREQUENCY AND AMPLITUDE; USED FOR HYPER TERMINAL
;-------------------------------------------------------------------------------------------------------------------------
Scale_Data
	movfw	Frequency				; set value in W register to Frequency
	movwf	Scaled_Frequency		; move value in W register to Scaled_Frequency
	movlw	H'02'
	movwf	Temp					; moves H'02' from W register to Temp

; changes the 8 bit frequency to 6 bit
Scale_Frequency
	rrf		Scaled_Frequency		; right shift Scaled_Frequency
	bcf		Scaled_Frequency,7		; clear bit 7
	decfsz	Temp					; decrements Temp and checks if 0...
	goto	Scale_Frequency			; if it's 0 then skip goto


	movfw	Amplitude				; set value in W register to Amplitude
	movwf	Scaled_Amplitude		; move value in W register to Scaled_Amplitude
	movlw	H'04'
	movwf	Temp					; moves H'04' from W register to Temp

; changes the 8 bit amplitude to 4 bit
Scale_Amplitude
	rrf		Scaled_Amplitude		; right shift Scaled_Amplitude
	bcf		Scaled_Amplitude,7		; clear bit 7
	decfsz	Temp					; decrements Temp and checks if 0...
	goto	Scale_Amplitude			; if it's 0 then skip goto

	return


;-------------------------------------------------------------------------------------------------------------------------
; Send_to_HyperTerminal - SENDS DATA TO HYPER TERMINAL
;-------------------------------------------------------------------------------------------------------------------------
Send_to_HyperTerminal

; move cursor home (top-left corner of HyperTerminal)
	movlw	H'1B'					; moves ascii char to W reg :	ESC
	call	Send_Char
	movlw	H'5B'					;								[
	call	Send_Char
	movlw	H'48'					;								H
	call	Send_Char

; move cursor down
	movlw	H'1B'					;								ESC
	call	Send_Char
	movlw	H'5B'					;								[
	call	Send_Char
	movlw	H'31'					;								1
	call	Send_Char
	movlw	H'38'					;								8
	call	Send_Char
	movlw	H'42'					;								B
	call	Send_Char
	
	movfw	Old_Scaled_Frequency
	movwf	Temp
	movlw	.1
	addwf	Temp
	
; move cursor forward
Move_Forward
	movlw	H'1B'					;								ESC
	call	Send_Char
	movlw	H'5B'					;								[
	call	Send_Char
	movlw	H'43'					;								C
	call	Send_Char
	decfsz	Temp
	goto	Move_Forward

	movfw	Old_Scaled_Amplitude
	movwf	Temp
	movlw	.1
	addwf	Temp
	
; erase current value
Erase_Value
	movlw	' '						;								' '
	call	Send_Char
	movlw	H'08'					;								backspace
	call	Send_Char

; move cursor up
	movlw	H'1B'					;								ESC
	call	Send_Char
	movlw	H'5B'					;								[
	call	Send_Char
	movlw	H'41'					;								A
	call	Send_Char

	decfsz	Temp
	goto	Erase_Value

; move cursor home
	movlw	H'1B'					; moves ascii char to W reg :	ESC
	call	Send_Char
	movlw	H'5B'					;								[
	call	Send_Char
	movlw	H'48'					;								H
	call	Send_Char

; move cursor down
	movlw	H'1B'					;								ESC
	call	Send_Char
	movlw	H'5B'					;								[
	call	Send_Char
	movlw	H'31'					;								1
	call	Send_Char
	movlw	H'38'					;								8
	call	Send_Char
	movlw	H'42'					;								B
	call	Send_Char
	
	movfw	Scaled_Frequency
	movwf	Temp
	movlw	.1
	addwf	Temp

; move cursor forward
Move_Forward_New
	movlw	H'1B'					;								ESC
	call	Send_Char
	movlw	H'5B'					;								[
	call	Send_Char
	movlw	H'43'					;								C
	call	Send_Char
	decfsz	Temp
	goto	Move_Forward_New

	movfw	Scaled_Amplitude
	movwf	Temp
	movlw	.1
	addwf	Temp

; send '*' char
Add_Star
	movlw	H'2A'					;								*
	call	Send_Char
	movlw	H'08'					;								backspace
	call	Send_Char

; move cursor up
Move_Up
	movlw	H'1B'					;								ESC
	call	Send_Char
	movlw	H'5B'					;								[
	call	Send_Char
	movlw	H'41'					;								A
	call	Send_Char

	decfsz	Temp
	goto	Add_Star

	return


;-------------------------------------------------------------------------------------------------------------------------
; Has_Scaled_Data_Changed - CHECKS IF (SCALED) FREQUENCY/AMPLITUDE CHANGED
;-------------------------------------------------------------------------------------------------------------------------
Has_Scaled_Data_Changed
	movfw	Scaled_Frequency			; move Scaled_Frequency to W register
	subwf	Old_Scaled_Frequency,W		; subtract Scaled_Frequency from Old_Scaled_Frequency and store in W
	btfss	STATUS,Z					; if W=0, STATUS=1, skip next operation
	bsf		Flags,SCALED_VALUE_CHANGED

	movfw	Scaled_Amplitude			; move Scaled_Amplitude to W register
	subwf	Old_Scaled_Amplitude,W		; subtract Scaled_Amplitude from Old_Scaled_Amplitude and store in W
	btfss	STATUS,Z					; if W=0, STATUS=1, skip next operation
	bsf		Flags,SCALED_VALUE_CHANGED

	return

	END
